



if not exist %LIBRARY_INC%\event2 exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist "%LIBRARY_PREFIX%/lib/event.lib" exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist "%LIBRARY_PREFIX%/lib/event_core.lib" exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist "%LIBRARY_PREFIX%/lib/event_extra.lib" exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
python event_rpcgen.py test/regress.rpc test/regress.gen.h test/regress.gen.c
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
